library UnitHistory requires optional AIDS, optional AutoIndex
	//===================================================================
	// Configurables.
	globals
		private constant real PERIOD = 0.2					//How frequently to save history steps
	endglobals
	
	private function UnitHistoryFilter takes unit u returns boolean
		return GetUnitAbilityLevel(u, 'ALoc') == 0
	endfunction
	
	//===================================================================
	// System code follows:
	//===================================================================
	// Utility function to convert time into a history step.
	public function ToSteps takes real time returns integer
		return R2I(time/PERIOD + 0.5)
	endfunction
	
	//===================================================================
	// A hashtable, for convenience.
	globals													//If only Table had RealTable and BooleanTable types
		private integer KEY = 0
		public hashtable HASH = InitHashtable()
	endglobals
	public function GetKey takes nothing returns integer		//Generate a new key for hashtable indexing
		set KEY = KEY + 1									//Hashtables don't really need index recycling
		return KEY
	endfunction
	
	//===================================================================
	// The integrated timer system.
	globals
		private trigger HistoryPeriodic = CreateTrigger()		//Has every UnitHistory addon's periodic method as a triggercondition
	endglobals
	
	private struct UnitHistoryTimer extends array				//Because I couldn't find an existing timer system that was well-suited
		timer Timer
		private integer TimerHandle							//Lol, efficiency?
		private boolean initialized							//Recycles timers when the autoindexer recycles indices
		static integer whichIndex								//"GetData()" - I like static members more than global variables, and it's private, anyway
		
		private static method periodic takes nothing returns nothing
			set thistype.whichIndex = LoadInteger(HASH, -1, GetHandleId(GetExpiredTimer()))
			call TriggerEvaluate(HistoryPeriodic)
		endmethod
		
		method start takes nothing returns nothing
			if not this.initialized then
				set this.Timer = CreateTimer()
				set this.TimerHandle = GetHandleId(this.Timer)
				set this.initialized = true
			endif
			call SaveInteger(HASH, -1, this.TimerHandle, this)
			call TimerStart(this.Timer, PERIOD, true, function thistype.periodic)
		endmethod
		
		method stop takes nothing returns nothing
			call PauseTimer(this.Timer)
			call RemoveSavedInteger(HASH, -1, this.TimerHandle) //Maybe not necessary, but safe
		endmethod
	endstruct
	
	//===================================================================
	// The struct to start/stop unit history.
	globals
		private trigger HistoryCreate = CreateTrigger()		//Has every UnitHistory addon's creation method as a triggercondition
		private trigger HistoryRemove = CreateTrigger()		//Has every UnitHistory addon's destroy method as a triggercondition
	endglobals
	
	struct UnitHistory										//Controls most automation, starting/stopping UnitHistory tracking on unit creation/removal
		private integer thisIndex								//Because AIDS can't get a unit's index during its onDestroy
		readonly static unit whichUnit						//Globals to pass data
		readonly static integer whichIndex
		
		readonly boolean enabled								//Is set to true if the unit passes the UnitHistory filter, therefore enabling history
		
		method toSteps takes real time returns integer			//More accurate than the public function since it considers the time since last save when rounding
			return R2I((time - TimerGetElapsed(UnitHistoryTimer[GetUnitId(whichUnit)].Timer))/PERIOD + 0.5)
		endmethod
		
		//===================================================================
		private integer lockLevel
		
		method lock takes nothing returns nothing				//Adds a lock to the unit; restore and trim will be disallowed for all UnitHistory addons
			if enabled then									//I wonder if it's more important to protect users from themselves, or to save an "if" call
				set lockLevel = lockLevel + 1
			endif
		endmethod
		
		method unlock takes nothing returns boolean			//Returns true if the unit history ended up unlocked
			if enabled and this.isLocked() then
				set lockLevel = lockLevel - 1
				return not this.isLocked()
			endif
			return enabled
		endmethod
		
		method isLocked takes nothing returns boolean
			return enabled and lockLevel > 0
		endmethod
		
		//===================================================================
		private integer suspendLevel
		
		method suspend takes nothing returns nothing			//Suspends the saving of further history steps for all UnitHistory addons for this unit
			if enabled then
				set suspendLevel = suspendLevel + 1
			endif
		endmethod
		
		method unsuspend takes nothing returns boolean			//Returns true if the unit history ended up unsuspended
			if enabled and this.isSuspended() then
				set suspendLevel = suspendLevel - 1
				return not this.isSuspended()
			endif
			return enabled
		endmethod
		
		method isSuspended takes nothing returns boolean
			return enabled and suspendLevel > 0
		endmethod
	
		//===================================================================
		//! runtextmacro optional AIDS()
		
		static if LIBRARY_AIDS then
			private method AIDS_onCreate takes nothing returns nothing
				set whichUnit = this.unit
		elseif LIBRARY_AutoIndex then
			private method onCreate takes nothing returns nothing
				set whichUnit = this.me
		endif
			if UnitHistoryFilter(whichUnit) then
				set enabled = true
				set suspendLevel = 0
				set lockLevel = 0
				
				set whichIndex = GetUnitId(whichUnit)
				set thisIndex = whichIndex
				call TriggerEvaluate(HistoryCreate)
				call UnitHistoryTimer[thisIndex].start()
			endif
		endmethod
		
		static if LIBRARY_AIDS then
			private method AIDS_onDestroy takes nothing returns nothing
		elseif LIBRARY_AutoIndex then
			private method onDestroy takes nothing returns nothing
		endif
			if enabled then
				set whichIndex = thisIndex
				call TriggerEvaluate(HistoryRemove)
				call UnitHistoryTimer[thisIndex].stop()
				set thisIndex = 0
				set suspendLevel = 0
				set lockLevel = 0
				set enabled = false
			endif
		endmethod
		
		implement optional AutoCreate
		implement optional AutoDestroy
	endstruct
	
	//===================================================================
	// The module with most of the functions provided by UnitHistory structs.
	module UnitHistory
		method toSteps takes real time returns integer			//More accurate than the public function since it the time since last save when rounding
			return whichUnit:UnitHistory.toSteps(time)
		endmethod
		
		//===================================================================
		// Public methods for people to call in order to utilize UnitHistory addons.
		readonly boolean enabled = false
		readonly integer currentStep = 0
		readonly integer whichStep = 0
		
		method step takes integer steps returns thistype		//Sets the current step for subsequent operations to a value a certain # of steps in the past
			if steps <= 0 then								//Can't go into the future
				set whichStep = currentStep					//If you try, clamp to the most recent step
			else
				if steps >= currentStep then					//Can't go back further than start of the history
					set whichStep = 0						//If you try, clamp to the oldest state
				else	
					set whichStep = currentStep - steps
				endif
			endif
			return this										//So you can chain method calls
		endmethod
		
		method time takes real time returns thistype			//Converts time to steps and sets the value for subsequent operations
			return this.step(.toSteps(time))
		endmethod
		
		method age takes nothing returns thistype				//Swaps from time-from-now (backtracking) to time-from-spawn (unit age)
			set whichStep = currentStep - whichStep
			return this
		endmethod
	
		method check takes integer whichStep returns boolean	//Returns false if the history does not contain the specified step
			return currentStep >= whichStep and whichStep >= 0
		endmethod
		
		method restore takes nothing returns thistype			//Restores a unit to the state selected using the functions above
			if enabled and not this.isLocked() and not whichUnit:UnitHistory.isLocked() then
				call this.revert()
			endif
			return this
		endmethod
		
		method trim takes nothing returns thistype				//Trims all saved history from the present up to but excluding the step specified
			if enabled and not this.isLocked() and not whichUnit:UnitHistory.isLocked() then
				if currentStep > whichStep then				//Do nothing if there is no history newer than the selected step
					if this.remove(currentStep) then			//If the addon specifies a need to destroy history states (eg. RemoveLocation),
						loop									//perform it for each trimmed step
							set currentStep = currentStep - 1
							exitwhen currentStep == whichStep
							call this.remove(currentStep)
						endloop
					else
						set currentStep = whichStep			//Otherwise, just change the index of the most recent step
						call this.onTrim()
					endif
				elseif currentStep == 0 then					//Special case: if only the initial history step remains, clear it
					call this.remove(currentStep)
					call this.onTrim()
					if not this.save() then					//Immediately save a new history state
						set currentStep = -1					//Revert the increment if the save was prevented
					endif
				endif
			endif
			return this
		endmethod
		
		method restoreTime takes real time returns nothing		//BJish because some people might prefer this syntax
			if time > 0 then
				call this.time(time).restore()
			endif
		endmethod
		
		method trimTime takes real time returns nothing		//BJish because some people might prefer this syntax
			if time > 0 then
				call this.time(time).trim()
			endif
		endmethod
		
		//===================================================================
		private integer lockLevel = 0
		
		method lock takes nothing returns nothing				//Adds a lock to this specific UnitHistory addon on the unit; restore and trim will be disallowed
			if enabled then
				set lockLevel = lockLevel + 1
			endif
		endmethod
		method unlock takes nothing returns boolean			//Returns true if the unit history ended up unlocked
			if enabled and this.isLocked() then
				set lockLevel = lockLevel - 1
				return not this.isLocked()
			endif
			return enabled
		endmethod
		method isLocked takes nothing returns boolean
			return enabled and lockLevel > 0
		endmethod
		
		//===================================================================
		private integer suspendLevel = 0
		
		method suspend takes nothing returns nothing			//Suspends the saving of further history steps for this UnitHistory addon for the unit
			if enabled then
				set suspendLevel = suspendLevel + 1
			endif
		endmethod
		method unsuspend takes nothing returns boolean			//Returns true if the unit history ended up unsuspended
			if enabled and this.isSuspended() then
				set suspendLevel = suspendLevel - 1
				return not this.isSuspended()
			endif
			return enabled
		endmethod
		method isSuspended takes nothing returns boolean
			return enabled and suspendLevel > 0
		endmethod
		
		//===================================================================
		// Administrative methods for UnitHistory addons.
		private static thistype array unitLink
		
		static method operator [] takes unit u returns thistype
			return unitLink[GetUnitId(u)]
		endmethod
		
		private static method periodic takes nothing returns boolean
			local thistype this = unitLink[UnitHistoryTimer.whichIndex]		//Load struct using index passed from the timer struct
			if this.enabled and not this.isSuspended() and not whichUnit:UnitHistory.isSuspended() then
				set this.currentStep = this.currentStep + 1
				if not this.save() then									//Save a history step
					set this.currentStep = this.currentStep - 1			//Revert the increment if the save was prevented
				endif
			endif
			return false
		endmethod
		
		method enable takes nothing returns boolean						//You might call this directly if you think a unit has changed and now passes the check
			if not enabled and thistype.filter(whichUnit) then			//Condition to actiavte this UnitHistory addon for this unit
				set enabled = true
				call this.init()											//Optionally specified by the add-on
				set currentStep = 0
				if not this.save() then									//Save the first history step right away
					set currentStep = -1									//Revert the increment if the save was prevented
				endif
			endif
			return enabled
		endmethod
		
		private static method onCreation takes nothing returns boolean
			local thistype this
			set this = thistype.create()
			set this.whichUnit = UnitHistory.whichUnit						//Retrieve which unit from the control struct
			set unitLink[UnitHistory.whichIndex] = this					//Link this struct to the unit index
			call this.enable()
			return false
		endmethod
		
		private static method onRemoval takes nothing returns boolean
			local thistype this = unitLink[UnitHistory.whichIndex]
			if this > 0 then
				call this.clear()
				set whichUnit = null
				set suspendLevel = 0
				set lockLevel = 0
				set enabled = false
				call this.destroy()
				set unitLink[UnitHistory.whichIndex] = 0
			endif
			return false
		endmethod
		
		private static method onInit takes nothing returns nothing								//Set up the methods to be run by UnitHistory
			call TriggerAddCondition(HistoryCreate, Condition(function thistype.onCreation))		//Save this addon's creation methods on the creation trigger
			call TriggerAddCondition(HistoryPeriodic, Condition(function thistype.periodic))		//Save this addon's periodic method on the creation trigger
			call TriggerAddCondition(HistoryRemove, Condition(function thistype.onRemoval))		//Save this addon's removal methods on the creation trigger
		endmethod
	endmodule
	
	//===================================================================
	// Null functions to fall back upon if not specified in a given addon.
	public struct Defaults
		method restore takes nothing returns thistype
			return this
		endmethod
		
		method save takes nothing returns boolean
			return true
		endmethod
		
		method revert takes nothing returns nothing
		endmethod
		
		method onTrim takes nothing returns nothing
		endmethod
		
		method remove takes integer whichStep returns boolean
			return false
		endmethod
		
		method clear takes nothing returns nothing
		endmethod
		
		static method filter takes unit u returns boolean
			return u != null
		endmethod
		
		method init takes nothing returns nothing
		endmethod
	endstruct
	
	//===================================================================
	// Call this to start a UnitHistory addon.
	//! textmacro UnitHistory takes IDENTIFIER
	struct $IDENTIFIER$History
		private static delegate UnitHistory_Defaults UnitHistory_Delegate = 0	//Fun times
		
		private unit whichUnit
		
		private static method hash takes nothing returns hashtable				//Just for convenience when writing UnitHistory add-ons
			return UnitHistory_HASH
		endmethod
		
		private static method newKey takes nothing returns integer				//Just for convenience when writing UnitHistory add-ons
			return UnitHistory_GetKey()
		endmethod
		
		implement UnitHistory
	//! endtextmacro
	
	//===================================================================
	// Call this to end a UnitHistory addon.
	//! textmacro endUnitHistory
	endstruct
	//! endtextmacro
endlibrary
